import json
import hooks._error_handlers
import hooks._settings
import hooks._logs
from log_trace.decorators import trace


@trace
def add_hooks(app):
    app.on_post_GET += _post_GET
    app.on_post_POST += _post_POST

    hooks._error_handlers.add_hooks(app)
    hooks._settings.add_hooks(app)
    hooks._logs.add_hooks(app)
    
    
def _post_POST(resource, request, payload):
    if payload.status_code == 201:
        j = json.loads(payload.data)
        if '_items' in j:
            for item in j['_items']:
                _edit_collection_link(request, item)
        else:
            _edit_collection_link(request, j)
        payload.data = json.dumps(j)


def _post_GET(resource, request, payload):
    if payload.status_code == 200:
        j = json.loads(payload.data)
        if '_items' in j:
            for item in j['_items']:
                _remove_unnecessary_links(item)
        else:
            _remove_unnecessary_links(j)
        payload.data = json.dumps(j)


def _remove_unnecessary_links(item):
    if 'related' in item['_links']:
        del item['_links']['related']


def _edit_collection_link(request, item):
    _remove_unnecessary_links(item)

    item['_links']['collection'] = {
        'href': request.path
    }



import os
import logging.config
import platform
import socket
import yaml
from eve import __version__ as eve_version
from cerberus import __version__ as cerberus_version
from werkzeug.utils import secure_filename

VERSION = '0.1.0'

# TODO: organize into functions!


def is_enabled(setting_key):
    setting_value = SETTINGS.get(setting_key)
    if not setting_value:
        return False
    return setting_value[0].lower() in 'yte'
    # i.e. the following means setting is enabled:
    # - 'Yes' or 'yes' or 'Y' or 'y'
    # - 'True' or 'true' or 'T' or 't'
    # - 'Enabled' or 'enabled' or 'E' or 'e'
    # TODO: refactor so lower level configurations can use (e.g. auth)


def set_optional_setting(var):
    if os.environ.get(var):
        SETTINGS[var] = os.environ.get(var)


def envar_to_int(envar, default=0):
    rtn = default
    try:
        rtn = int(os.environ.get(envar, default))
    except ValueError:
        rtn = default

    return rtn



# set environment variables from _env.conf (which is in .gitignore)
if os.path.exists('_env.conf'):
    with open('_env.conf') as setting:
        for line in setting:
            if not line.startswith('#'):
                line = line.rstrip()
                nvp = line.split('=')
                if len(nvp) == 2:
                    os.environ[nvp[0].strip()] = nvp[1].strip()


# TODO: sanitize smtp/email_recipients
SETTINGS = {
    'ES_API_NAME': '{$project_name}',

    'ES_MONGO_ATLAS': os.environ.get('ES_MONGO_ATLAS', 'Disabled'),
    'ES_MONGO_HOST': os.environ.get('ES_MONGO_HOST', 'localhost'),
    'ES_MONGO_PORT': envar_to_int('ES_MONGO_PORT', 27017),
    'ES_MONGO_DBNAME': os.environ.get('ES_MONGO_DBNAME', '{$project_name}'),
    'ES_API_PORT': envar_to_int('ES_API_PORT', 2112),
    'ES_INSTANCE_NAME': os.environ.get('ES_INSTANCE_NAME', socket.gethostname()),
    'ES_TRACE_LOGGING': os.environ.get('ES_TRACE_LOGGING', 'Enabled'),
    'ES_PAGINATION_LIMIT': envar_to_int('ES_PAGINATION_LIMIT', 3000),
    'ES_PAGINATION_DEFAULT': envar_to_int('ES_PAGINATION_DEFAULT', 1000),
    'ES_LOG_TO_FOLDER': os.environ.get('ES_LOG_TO_FOLDER', 'Enabled'),
    'ES_SEND_ERROR_EMAILS': os.environ.get('ES_SEND_ERROR_EMAILS', 'Enabled'),
    'ES_SMTP_HOST': os.environ.get('ES_SMTP_HOST', 'internalmail.cri.com'),
    'ES_SMTP_PORT': envar_to_int('ES_SMTP_PORT', 25),
    'ES_ERROR_EMAIL_RECIPIENTS': os.environ.get('ES_ERROR_EMAIL_RECIPIENTS', 'michael@pointw.com')
}

# optional settings...
set_optional_setting('ES_MONGO_USERNAME')
set_optional_setting('ES_MONGO_PASSWORD')
set_optional_setting('ES_MONGO_AUTH_SOURCE')
set_optional_setting('ES_MEDIA_BASE_URL')
set_optional_setting('ES_PUBLIC_RESOURCES')

# cancellable settings...
# if SETTINGS.get('ES_CANCELLABLE') == '':
#     del SETTINGS['ES_CANCELLABLE']


# Set up logging
API_NAME = SETTINGS.get('ES_API_NAME')

with open('logging.yml', 'r') as f:
    logging_config = yaml.load(f, Loader=yaml.FullLoader)

if is_enabled('ES_LOG_TO_FOLDER'):
    LOG_FOLDER = f'/var/log/{secure_filename(API_NAME)}'
    if not os.path.exists(LOG_FOLDER):
        os.makedirs(LOG_FOLDER)

    logging_config['handlers']['all']['filename'] = os.path.join(LOG_FOLDER, 'all.log')
    logging_config['handlers']['warn']['filename'] = os.path.join(LOG_FOLDER, 'warn.log')
    logging_config['handlers']['error']['filename'] = os.path.join(LOG_FOLDER, 'error.log')

if is_enabled('ES_SEND_ERROR_EMAILS'):
    logging_config['handlers']['smtp']['mailhost'] = [SETTINGS.get('ES_SMTP_HOST'), SETTINGS.get('ES_SMTP_PORT')]
    logging_config['handlers']['smtp']['toaddrs'] = [e.strip() for e in SETTINGS.get('ES_ERROR_EMAIL_RECIPIENTS').split(',')]
    logging_config['handlers']['smtp']['subject'] = f'Problem encountered with {API_NAME}'


logging.config.dictConfig(logging_config)

werkzeug_log = logging.getLogger('werkzeug')
werkzeug_log.setLevel(logging.ERROR)

LOG = logging.getLogger('configuration')
LOG.info('%s version:  %s', API_NAME, VERSION)
LOG.info('Eve version:      %s', eve_version)
LOG.info('Cerberus version: %s', cerberus_version)
LOG.info('Python version:   %s', platform.sys.version)



if is_enabled('ES_SEND_ERROR_EMAILS'):
    instance_name = SETTINGS.get('ES_INSTANCE_NAME')
    EMAIL_FORMAT = f'''%(levelname)s sent from {API_NAME} instance "{instance_name}" (hostname: {socket.gethostname()})

    %(asctime)s - %(levelname)s - File: %(filename)s - %(funcName)s() - Line: %(lineno)d -  %(message)s
    '''

    EMAIL_FORMAT += f'''
    {API_NAME} version:  {VERSION}
    Eve version:      {eve_version}
    Cerberus version: {cerberus_version}
    Python version:   {platform.sys.version}

    '''

    for setting in sorted(SETTINGS):
        key = setting.upper()
        if ('PASSWORD' not in key) and ('SECRET' not in key):
            LOG.info(f'{setting}: {SETTINGS[setting]}')
            EMAIL_FORMAT += f'{setting}: {SETTINGS[setting]}\n'
    EMAIL_FORMAT += '\n\n'

    LOGGER = logging.getLogger()
    HANDLERS = LOGGER.handlers

    SMTP_HANDLER = [x for x in HANDLERS if x.name == 'smtp'][0]
    SMTP_HANDLER.setFormatter(logging.Formatter(EMAIL_FORMAT))



"""
Settings to configure Eve's behaviours.
"""
import domain
from configuration import SETTINGS, is_enabled

if is_enabled('ES_MONGO_ATLAS'):
    MONGO_URI = f'mongodb+srv://{SETTINGS.get("ES_MONGO_USERNAME")}:{SETTINGS.get("ES_MONGO_PASSWORD")}@{SETTINGS["ES_MONGO_HOST"]}/{SETTINGS["ES_MONGO_DBNAME"]}?retryWrites=true&w=majority'
else:
    MONGO_HOST = SETTINGS.get('ES_MONGO_HOST')
    MONGO_PORT = SETTINGS.get('ES_MONGO_PORT')
    MONGO_DBNAME = SETTINGS.get('ES_MONGO_DBNAME')
    if 'ES_MONGO_AUTH_SOURCE' in SETTINGS.keys():
        MONGO_AUTH_SOURCE = SETTINGS.get('ES_MONGO_AUTH_SOURCE')
    if 'ES_MONGO_USERNAME' in SETTINGS.keys():
        MONGO_USERNAME = SETTINGS.get('ES_MONGO_USERNAME')
    if 'ES_MONGO_PASSWORD' in SETTINGS.keys():
        MONGO_PASSWORD = SETTINGS.get('ES_MONGO_PASSWORD')

# the default BLACKLIST is ['$where', '$regex'] - the following line turns on regex
MONGO_QUERY_BLACKLIST = ['$where']

RENDERERS = ['eve.render.JSONRenderer']  # removed eve.render.XMLRenderer

DATE_FORMAT = '%Y-%m-%dT%H:%M:%S'
PAGINATION_LIMIT = SETTINGS.get('ES_PAGINATION_LIMIT')
PAGINATION_DEFAULT = SETTINGS.get('ES_PAGINATION_DEFAULT')
OPTIMIZE_PAGINATION_FOR_SPEED = False


# http://python-eve.org/features.html#operations-log
# OPLOG = True
# OPLOG_ENDPOINT = '_oplog'

SCHEMA_ENDPOINT = '_schema'
RESOURCE_METHODS = ['GET', 'POST', 'DELETE']
ITEM_METHODS = ['GET', 'PATCH', 'DELETE']

X_DOMAINS = '*'
X_EXPOSE_HEADERS = ['Origin', 'X-Requested-With', 'Content-Type', 'Accept']
X_HEADERS = [
    'Accept',
    'Authorization',
    'If-Match',
    'Access-Control-Expose-Headers',
    'Access-Control-Allow-Origin',
    'Content-Type',
    'Pragma',
    'X-Requested-With',
    'Cache-Control'
]

RETURN_MEDIA_AS_BASE64_STRING = False
RETURN_MEDIA_AS_URL = True

if 'ES_MEDIA_BASE_URL' in SETTINGS.keys():
    MEDIA_BASE_URL = SETTINGS.get('ES_MEDIA_BASE_URL')
EXTENDED_MEDIA_INFO = ['content_type', 'name', 'length']

DOMAIN = domain.DOMAIN

UPLOAD_FOLDER = 'uploads/'



"""
hooks.logs
This module defines functions to log requests, and to manage log verbosity.
"""
import logging
import json
from flask import abort, make_response, jsonify, request as flask_request
from log_trace.decorators import trace
from utils import make_error_response

LOG = logging.getLogger('hooks.logs')


@trace
def add_hooks(app):
    """Wire up the events for logging"""
    app.on_post_GET += _log_request
    app.on_post_POST += _log_request
    app.on_post_PATCH += _log_request
    app.on_post_PUT += _log_request
    app.on_post_DELETE += _log_request

    @app.route('/_logging', methods=['GET'])
    def get_logging_config():
        """Returns the current verbosity levels for logging handlers."""
        if app.auth and not app.auth.authorized(None, '_logging', 'GET'):
            return make_error_response('Please provide proper credentials', 401)

        return _get_logging_config()

    @app.route('/_logging', methods=['PUT'])
    def put_logging_config():
        """PUT logging level to handlers."""
        if app.auth and not app.auth.authorized(None, '_logging', 'PUT'):
            return make_error_response('Please provide proper credentials', 401)

        return _put_logging_config(flask_request)


@trace
def _log_request(resource, request, payload):
    """Event hook to log all requests."""
    LOG.info(f'Request for {resource}: {request} [{payload.status_code}]')
    LOG.debug(f'Request for {resource}: {request} [{payload.status_code}] '
              f'{request.values} {payload.data} [{request.headers}]')


@trace
def _get_logging_config():
    """Returns the verbosity for all handlers."""
    logger = logging.getLogger()
    payload = {
        handler.name: logging.getLevelName(handler.level)
        for handler in logger.handlers
    }

    response = make_response(jsonify(payload), 200)
    _log_request('_logging', request, response)
    return response


@trace
def _put_logging_config(request):
    """PUTs the verbosity for handlers."""
    response = make_error_response('Could not change log settings', 400)

    try:
        if request.content_type != 'application/json':
            raise TypeError('The request body must be application/json')

        payload = json.loads(request.data)

        logger = logging.getLogger()
        # first loop through to ensure everything is valid
        for key in payload:
            handler = [x for x in logger.handlers if x.name == key]
            if not handler:
                raise ValueError(f'{key} is not a valid log handler')
            try:
                getattr(logging, payload[key])
            except AttributeError:
                raise ValueError(f'{payload[key]} is not a valid log verbosity level')

        # now it's safe to iterate and change levels
        for key in payload:
            handler = [x for x in logger.handlers if x.name == key][0]
            handler.setLevel(getattr(logging, payload[key]))

        payload = {}
        for handler in logger.handlers:
            payload[handler.name] = logging.getLevelName(handler.level)

        response = make_response(jsonify(payload), 200)

    except (TypeError, ValueError) as ex:
        response = make_error_response('Invalid log setting specification', 422, exception=ex)

    except Exception as ex:
        response = make_error_response('Could not change log settings', 400, exception=ex)

    _log_request('_logging', request, response)
    return response



import logging
from eve import Eve
from flask_cors import CORS
import hooks
from configuration import SETTINGS

LOG = logging.getLogger('{$project_name}')

name = SETTINGS.get('ES_API_NAME', '{$project_name}')
app = Eve(import_name=name)
CORS(app)
hooks.add_hooks(app)

LOG.info('Starting serverless')



from pymongo.errors import ServerSelectionTimeoutError
from utils import make_error_response


def add_hooks(app):
    @app.errorhandler(Exception)
    def catchall_exception_handler(ex):
        return make_error_response('Unexpected server error', 500, exception=ex)

    @app.errorhandler(ServerSelectionTimeoutError)
    def mongo_missing_exception_handler(ex):
        message = 'Could not connect to mongodb.  ' \
                  'Ask server admin to check connection settings and ensure database is running.'
        return make_error_response(message, 500, exception=ex)



"""
Defines the resources that comprise the {$project_name} domain.
"""
from . import _settings


DOMAIN_DEFINITIONS = {
    '_settings': _settings.DEFINITION
}


DOMAIN_RELATIONS = {
}


DOMAIN = {**DOMAIN_DEFINITIONS, **DOMAIN_RELATIONS}



import win32serviceutil
import win32service
import win32event
import servicemanager

from eve_service import EveService


class EveWindowsService(win32serviceutil.ServiceFramework):
    _svc_name_ = "{$project_name}"
    _svc_display_name_ = "{$project_name} API"
    # _svc_description_ = "Service description goes here."  # TODO: replace with config or some such

    def __init__(self, args):
        self._eve = EveService()

        win32serviceutil.ServiceFramework.__init__(self, args)
        self.hWaitStop = win32event.CreateEvent(None, 0, 0, None)

    def SvcDoRun(self):
        servicemanager.LogMsg(servicemanager.EVENTLOG_INFORMATION_TYPE, servicemanager.PYS_SERVICE_STARTED,
                              (self._svc_name_, ''))

        self._eve.start()

        servicemanager.LogMsg(servicemanager.EVENTLOG_INFORMATION_TYPE, servicemanager.PYS_SERVICE_STOPPED,
                              (self._svc_name_, ''))

    def SvcStop(self):
        self._eve.stop()
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
        win32event.SetEvent(self.hWaitStop)


if __name__ == '__main__':
    win32serviceutil.HandleCommandLine(EveWindowsService)



"""
Defines custom validations used in domain.
"""
import re
import logging
import isodate

from eve.utils import config
from eve.io.mongo import Validator
from bson.objectid import ObjectId

from utils import get_db
from log_trace.decorators import trace


LOG = logging.getLogger('validator')


class EveValidator(Validator):
    """Validator for custom types and validations."""
    @trace
    def _validate_unique_ignorecase(self, unique_ignorecase, field, value):
        """ Validates that a field value is unique, ignoring case.
            NOTE: this method was copy/pasted from Eve io/mongo/validation, then made
                  case insensitive

        The rule's arguments are validated against this schema:
        {'type': 'boolean'}
        """
        query = {}
        if unique_ignorecase:
            query = {
                field: re.compile('^' + re.escape(value) + '$', re.IGNORECASE)
            }

            resource_config = config.DOMAIN[self.resource]

            # exclude soft deleted documents if applicable
            if resource_config["soft_delete"]:
                # be aware that, should a previously (soft) deleted document be
                # restored, and because we explicitly ignore soft deleted
                # documents while validating 'unique' fields, there is a chance
                # that a unique field value will end up being now duplicated
                # in two documents: the restored one, and the one which has
                # been stored with the same field value while the original
                # document was in 'deleted' state.

                # we make sure to also include documents which are missing the
                # DELETED field. This happens when soft deletes are enabled on
                # an a resource with existing documents.
                query[config.DELETED] = {"$ne": True}

            # exclude current document
            if self.document_id:
                id_field = resource_config["id_field"]
                query[id_field] = {"$ne": self.document_id}

            # we perform the check on the native mongo driver (and not on
            # app.data.find_one()) because in this case we don't want the usual
            # (for eve) query injection to interfere with this validation. We
            # are still operating within eve's mongo namespace anyway.

            if get_db()[self.resource].find_one(query):
                self._error(field, "value '%s' is not unique (case-insensitive)" % value)

    @trace
    def _validate_unique_to_parent(self, unique_to_parent, field, value):
        """
        The rule's arguments are validated against this schema:
        {'type': 'string'}
        """
        if not unique_to_parent:
            return

        resource = self.resource
        rel = resource.split('_')
        if len(rel) > 1:
            resource = rel[1]

        parent_ref_field = f'_{unique_to_parent}_ref'  # TODO: must be singular for now - fix this

        # TODO: assert(parent_ref_field in self.schema.schema)
        # TODO: assert('data_relation' in self.schema.schema[parent_ref_field])
        parent_resource = self.schema.schema[parent_ref_field]['data_relation'].get('resource')
        # TODO: assert(parent_resource)
        parent_ref = self.document.get(parent_ref_field)

        collection = get_db()[resource]
        query = {
            field: re.compile('^' + re.escape(value) + '$', re.IGNORECASE)
        }

        query[parent_ref_field] = ObjectId(parent_ref) if parent_ref else None

        prior = collection.find_one(query)
        if prior and unique_to_parent:
            prior_name = prior.get('name')
            message = f'/{parent_resource}/{parent_ref}/{resource} already has an item whose {field} is {prior_name}'
            if not parent_ref:
                message = f'/{resource} already has an item whose {field} is {prior_name}'
            self._error(field, message)

    # ISO type definitions
    @trace
    def _validate_type_iso_date(self, date_value):
        is_valid = True
        try:
            isodate.parse_date(date_value)
            if not re.match(r'^([0-9]{4})-?((1[0-2]|0[1-9])-?(3[01]|0[1-9]|[12][0-9])|(W([0-4]\d|5[0-2])(-?[1-7]))|((00[1-9]|0[1-9]\d|[12]\d{2}|3([0-5]\d|6[1-6]))))$', date_value): # pylint: disable=line-too-long
                is_valid = False
        except isodate.ISO8601Error as ex:
            is_valid = False

        if is_valid:
            return True

    @trace
    def _validate_type_iso_time(self, time_value):
        is_valid = True
        if time_value == '24:00':
            return is_valid
        try:
            isodate.parse_time(time_value)
            if not re.match(r'^([01]\d|2[0-3])\D?([0-5]\d)\D?([0-5]\d)?\D?(\d{3})?$', time_value):
                is_valid = False
        except isodate.ISO8601Error as ex:
            is_valid = False

        if is_valid:
            return True

    @trace
    def _validate_type_iso_duration(self, duration_value):
        is_valid = True
        try:
            isodate.parse_duration(duration_value)
        except isodate.ISO8601Error as ex:
            is_valid = False

        if is_valid:
            return True



"""Home of the auth modules that govern access via Eve."""
import os
import logging
import jwt

LOG = logging.getLogger('auth')

SETTINGS = {
    'ES_AUTH_ADD_BASIC': os.environ.get('ES_AUTH_ADD_BASIC', 'No'),  # [0] in 'yYtT', i.e. yes, Yes, true, True
    'ES_AUTH_ROOT_PASSWORD': os.environ.get('ES_AUTH_ROOT_PASSWORD', 'password'),
    'ES_AUTH_REALM': os.environ.get('ES_AUTH_REALM', '{$project_name}.pointw.com'),

    'ES_AUTH_JWT_DOMAIN': os.environ.get('ES_AUTH_JWT_DOMAIN', '{$project_name}.us.auth0.com'),
    'ES_AUTH_JWT_ISSUER': os.environ.get('ES_AUTH_JWT_ISSUER', 'https://{$project_name}.us.auth0.com/'),
    'ES_AUTH_JWT_AUDIENCE': os.environ.get('ES_AUTH_JWT_AUDIENCE', 'https://pointw.com/{$project_name}'),

    'AUTH0_API_AUDIENCE': os.environ.get('AUTH0_API_AUDIENCE', 'https://{$project_name}.us.auth0.com/api/v2/'),
    'AUTH0_API_BASE_URL': os.environ.get('AUTH0_API_BASE_URL', 'https://{$project_name}.us.auth0.com/api/v2'),
    'AUTH0_CLAIMS_NAMESPACE': os.environ.get('AUTH0_CLAIMS_NAMESPACE', 'https://pointw.com/{$project_name}'),
    'AUTH0_TOKEN_ENDPOINT': os.environ.get('AUTH0_TOKEN_ENDPOINT', 'https://{$project_name}.us.auth0.com/oauth/token'),
    'AUTH0_CLIENT_ID': os.environ.get('AUTH0_CLIENT_ID', '--your-client-id--'),
    'AUTH0_CLIENT_SECRET': os.environ.get('AUTH0_CLIENT_SECRET',
                                          '--your-client-secret--')
}

JWK_CLIENT = jwt.PyJWKClient(f'https://{SETTINGS["ES_AUTH_JWT_DOMAIN"]}/.well-known/jwks.json')
_jwks = JWK_CLIENT.get_signing_keys()
SIGNING_KEYS = {jwk.key_id: jwk.key for jwk in _jwks}


# cancellable
if SETTINGS['ES_AUTH_JWT_AUDIENCE'] == '':
    del SETTINGS['ES_AUTH_JWT_AUDIENCE']

for setting in SETTINGS:
    key = setting.upper()
    if ('PASSWORD' not in key) and ('SECRET' not in key):
        LOG.info('%s: %s', setting, SETTINGS[setting])



"""
hooks.settings
This module defines functions used by the other hooks modules, and some hooks of its own.
"""
import logging
import platform
from flask import current_app, abort
from utils import make_error_response
from configuration import SETTINGS, VERSION
from eve import __version__ as eve_version
from cerberus import __version__ as cerberus_version
from log_trace.decorators import trace


LOG = logging.getLogger('hooks.utils')


@trace
def add_hooks(app):
    """Wire up the events for _settings endpoint."""
    app.on_fetched_resource__settings += _fetch_settings


@trace
def _fetch_settings(response):
    if current_app.auth and not current_app.auth.authorized(None, '_settings', 'GET'):
        abort(make_error_response('Please provide proper credentials', 401))

    del response['_items']
    del response['_meta']

    response['versions'] = {}
    response['settings'] = {}

    response['versions']['{$project_name}'] = VERSION
    response['versions']['eve'] = eve_version
    response['versions']['cerberus'] = cerberus_version
    response['versions']['python'] = platform.sys.version

    for env in sorted(SETTINGS):
        key = env.upper()
        if ('PASSWORD' not in key) and ('SECRET' not in key):
            response['settings'][env] = SETTINGS[env]



import logging
from flask import jsonify, make_response
from flask import current_app


LOG = logging.getLogger('utils')


def get_db():
    return current_app.data.driver.db


def make_error_response(message, code, issues=[], **kwargs):
    if 'exception' in kwargs:
        ex = kwargs.get('exception')
        LOG.exception(message, ex)

        if ex:
            issues.append({
                'exception': {
                    'name': type(ex).__name__,
                    'type': ".".join([type(ex).__module__, type(ex).__name__]),
                    'args': ex.args
                }
            })

    resp = {
        '_status': 'ERR',
        '_error': {
            'message': message,
            'code': code
        }
    }

    if issues:
        resp['_issues'] = issues

    return make_response(jsonify(resp), code)





import os
import signal
from eve_service import EveService


def stop_api(signum, frame):  # pylint: disable=unused-argument
    """ Catches SIGTERM and issues SIGINT """
    if signum == signal.SIGTERM:
        os.kill(os.getpid(), signal.SIGINT)


if __name__ == '__main__':
    signal.signal(signal.SIGTERM, stop_api)

    eve = EveService()

    try:
        eve.start()
    except KeyboardInterrupt:
        eve.stop()



import logging
from eve import Eve
from flask_cors import CORS
import hooks
from configuration import SETTINGS

LOG = logging.getLogger('{$project_name}')


class EveService:
    def __init__(self):
        self._name = SETTINGS.get('ES_API_NAME', '{$project_name}')
        self._app = Eve(import_name=self._name)
        CORS(self._app)
        hooks.add_hooks(self._app)

    def start(self):
        border = '-' * (23 + len(self._name))
        LOG.info(border)
        LOG.info(f'****** STARTING {self._name} ******')
        LOG.info(border)
        try:
            self._app.run(host='0.0.0.0', port=SETTINGS.get('ES_API_PORT'), threaded=True)
        except Exception as ex:  # pylint: disable=broad-except
            LOG.exception(ex)
        finally:
            LOG.info(border)
            LOG.info(f'****** STOPPING {self._name} ******')
            LOG.info(border)

    def stop(self):
        self._app.do_teardown_appcontext()



"""
Function decorators used in this project
- @trace - log_trace entries and exits of decorated function
"""
import logging
from functools import wraps
from typing import Any, Callable
from configuration import SETTINGS

LOG = logging.getLogger('trace')


def _trace(target):
    """
    This is a decorator that log_trace when the decorated function is entered and exited.
    Any unhandled exceptions raised by the decorated function are logged and re-raised.
    """

    @wraps(target)
    def _wrapper(*args, **kwargs):
        return_value = None
        LOG.trace(f'Entering function: {target.__module__}.{target.__name__}')

        try:
            return_value = target(*args, **kwargs)
        except Exception as ex:  # pylint: disable=broad-except
            LOG.trace(f'Exiting function: {target.__module__}.{target.__name__} by unhandled exception "{ex}"')
            LOG.exception(ex)
            raise

        LOG.trace(f'Exiting function: [{target.__module__}].[{target.__name__}] by return')

        return return_value

    return _wrapper


if SETTINGS.get('ES_TRACE_LOGGING') == 'Enabled':
    trace = _trace
else:
    trace: Callable[[Any], Any] = lambda fn: fn



import inflect

def get_pair(word):
    p = inflect.engine()
    
    singular = word if p.singular_noun(word) == False else p.singular_noun(word)
    plural = word if p.singular_noun(word) else p.plural_noun(word)
    
    if singular == plural:  # in case of words like moose, fish
        plural = plural + 's'

    return (singular, plural)


"""
The auth module used for ES.
"""
from bson.objectid import ObjectId
from eve_negotiable_auth import NegotiableAuth, AUTH_PARSER

from utils import get_db
from . import SETTINGS
from .auth_handlers import basic, bearer, bearer_challenge

AUTH_PARSER.add_handler('Bearer', bearer, bearer_challenge, realm=f'{SETTINGS["ES_AUTH_REALM"]}')
if SETTINGS['ES_AUTH_ADD_BASIC'][0] in 'tyTY':
    AUTH_PARSER.add_handler('Basic', basic, realm=f'{SETTINGS["ES_AUTH_REALM"]}')


class EveServiceAuth(NegotiableAuth):
    def __init__(self):
        super(EveServiceAuth, self).__init__()

    def process_claims(self, claims, allowed_roles, resource, method):
        authorized = 'user' in claims
        if not authorized:
            return False

        is_admin = claims.get('role') == 'admin'

        if not is_admin:
            auth_value = 'all-denied'
            accounts = get_db()['accounts']
            account = accounts.find_one({'user_id': claims['user']})
            if account:
                auth_value = ObjectId(account['_customer_ref'])
                claims['role'] = auth_value

            self.set_request_auth_value(auth_value)

        if resource in ['accounts', 'roles'] and not is_admin:
            authorized = False

        return authorized



"""
Methods to access/modify user information in Auth0
https://auth0.com/docs/api/management/v2
"""
import logging
import json
import requests
from log_trace.decorators import trace
from auth import SETTINGS

LOG = logging.getLogger('auth0')


@trace
def get_token():
    token_url = SETTINGS['AUTH0_TOKEN_ENDPOINT']
    client_id = SETTINGS['AUTH0_CLIENT_ID']
    client_secret = SETTINGS['AUTH0_CLIENT_SECRET']
    audience = SETTINGS['AUTH0_API_AUDIENCE']

    body = json.dumps({
        'client_id': client_id,
        'client_secret': client_secret,
        'audience': audience,
        'grant_type': 'client_credentials'
    })
    headers = {
        'content-type': 'application/json'
    }
    response = requests.post(token_url, headers=headers, data=body)
    # TODO: robustify
    token = response.json()['access_token']
    return token


@trace
def get_users(token):
    get_users_url = f'{SETTINGS["AUTH0_API_BASE_URL"]}/users'
    headers = {
        'Authorization': 'Bearer ' + token
    }
    response = requests.get(get_users_url, headers=headers)
    return response.json()


@trace
def get_user_roles(user_id, token):
    get_roles_url = f'{SETTINGS["AUTH0_API_BASE_URL"]}/users/{user_id}/roles'

    headers = {
        'Authorization': 'Bearer ' + token
    }
    response = requests.get(get_roles_url, headers=headers)
    # TODO: robustify
    return response.json()


@trace
def add_user_roles(token, user_id, roles_to_add, remove=False):
    # TODO: refactor ugly bad handling
    # NOTE: role names are case sensitive
    get_roles_url = f'{SETTINGS["AUTH0_API_BASE_URL"]}/roles'
    headers = {
        'Authorization': 'Bearer ' + token,
        'Content-type': 'application/json'
    }
    response = requests.get(get_roles_url, headers=headers)
    roles = response.json()
    role_ids = [role['id'] for role in roles if role['name'] in roles_to_add]

    request_body = {
        'roles': role_ids
    }
    if role_ids:
        user_roles_url = f'{SETTINGS["AUTH0_API_BASE_URL"]}/users/{user_id}/roles'
        if remove:
            response = requests.delete(user_roles_url, headers=headers, data=json.dumps(request_body))
        else:
            response = requests.post(user_roles_url, headers=headers, data=json.dumps(request_body))
        LOG.info(f'{"Removed" if remove else "Added"} roles on {user_id}: {response.status_code}')
        if response.status_code != 204:
            return False
    else:
        LOG.warning('no valid roles specified')
        return False

    return True


@trace
def delete_user(token, user_id):
    user_url = f'{SETTINGS["AUTH0_API_BASE_URL"]}/users/{user_id}'
    headers = {
        'Authorization': 'Bearer ' + token
    }
    response = requests.delete(user_url, headers=headers)

    LOG.info(f'"Removed {user_id} from auth0: {response.status_code}')
    return response.status_code == 204



"""Defines TRACE logging verbosity"""
import logging

TRACE_LEVEL = 5
logging.addLevelName(TRACE_LEVEL, 'TRACE')


def trace(self, message, *args, **kws):
    """Extends logging with TRACE level"""
    if self.isEnabledFor(TRACE_LEVEL):
        self._log(TRACE_LEVEL, message, args, **kws)  # pylint: disable=protected-access


logging.Logger.trace = trace



from glob import glob
from setuptools import setup

setup(
    name='eve-utils',
    version='0.9.10',
    description='Templates and scripts to rapidly spin up a production-ready Eve-based API.',
    long_description=open('../README.md').read(),
    long_description_content_type='text/markdown',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'Framework :: Flask',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Utilities'
    ],
    url='https://github.com/pointw-dev/eve-utils',
    author='Michael Ottoson',
    author_email='michael@pointw.com',
    packages=['eve_utils'],
    include_package_data=True,
    install_requires=[
        'libcst',
        'inflect'
    ],
    scripts=glob('bin/*'),
    zip_safe=False
)



import os
import sys
import subprocess
from distutils.dir_util import copy_tree, remove_tree
import eve_utils


def insert_import(original_body, addition):
    rtn = []
    state = 'on-top'
    for item in original_body:
        if state == 'on-top':
            if hasattr(item, 'body') and hasattr(item.body, '__iter__') and type(item.body[0]).__name__ in ['Import', 'ImportFrom', 'Expr']:
                pass
            else:
                state = 'in-position'

        if state == 'in-position':
            rtn.append(addition)  # TODO: if no other appends before, add newline after here
            state = 'on-bottom'

        rtn.append(item)

    return rtn


def install_packages(packages, command):
    trigger = 'Successfully installed '
    subprocess.check_output([sys.executable, "-m", "pip", "install", "-r", "requirements.txt"])

    with open('requirements.txt', 'a') as f:
        f.write(f'\n# start: added by {command}\n')
        for package in packages:
            out = subprocess.check_output([sys.executable, "-m", "pip", "install", package]).decode('utf-8')
            for line in out.split('\n'):
                if line.startswith(trigger):
                    installed_packages = line[len(trigger):].split(' ')

                    for installed_package in installed_packages:
                        if package in installed_package:
                            hyphen = installed_package.rfind('-')
                            f.write(f'{installed_package[:hyphen]}=={installed_package[hyphen+1:]}\n')

        f.write(f'# end: added by {command}\n')


def copy_skel(project_name, skel_folder, target_folder=None):
    print(f'Adding {skel_folder} to {project_name} API')

    source = os.path.join(os.path.dirname(eve_utils.__file__), f'skel/{skel_folder}')
    destination = skel_folder if not target_folder else target_folder

    if not target_folder:
        os.mkdir(skel_folder)  # TODO: ensure doesn't already exist, etc
    copy_tree(source, destination)

    # TODO: can the following remove_tree calls be obviated if skel is packaged differently?
    remove_if_exists(os.path.join(destination, '__pycache__'))

    for dname, dirs, files in os.walk(destination):
        for fname in files:
            fpath = os.path.join(dname, fname)
            if fname.endswith('.pyc') or fname.endswith('.ico'):
                continue
            with open(fpath, 'r') as f:
                try:
                    s = f.read()
                except UnicodeDecodeError as ex:
                    continue
            changed = False
            if '{$project_name}' in s:
                s = s.replace("{$project_name}", project_name)
                changed = True
            if changed:
                with open(fpath, "w") as f:
                    f.write(s)


def replace_project_name(project_name, folder=None):
    if not folder:
        folder = f'./{project_name}'
    for dname, dirs, files in os.walk(folder):
        for fname in files:
            # do not process if traversing nested venv folder
            if os.path.abspath(dname).startswith(sys.prefix):
              continue
            if '__pycache__' in dname:
              continue
            fpath = os.path.join(dname, fname)
            try:
              with open(fpath) as f:
                  s = f.read()
              s = s.replace("{$project_name}", project_name)
              with open(fpath, "w") as f:
                  f.write(s)
            except UnicodeDecodeError as ex:
              print(f'Skipping unprocessable file: {dname}/{fname}')



def remove_if_exists(folder):
    if os.path.exists(folder):
        remove_tree(folder)



import jwt
from base64 import b64decode
from auth import SETTINGS, SIGNING_KEYS


def basic(token, **kwargs):
    try:
        username, password = b64decode(token).decode().split(':', 1)
        is_root = False
        if username.lower() == 'root':
            if password == SETTINGS.get('ES_AUTH_ROOT_PASSWORD'):
                is_root = True
            else:
                return {}

        role = 'admin' if is_root else ''
        rtn = {
            'user': username,
            'role': role
        }
    except:
        rtn = {}

    return rtn


# TODO: detect opaque token, then handle accordingly (how?)
def bearer(token, **kwargs):
    audience = SETTINGS.get('ES_AUTH_JWT_AUDIENCE')
    issuer = SETTINGS.get('ES_AUTH_JWT_ISSUER')

    options = {}

    if audience:
        options['audience'] = audience
    if issuer:
        options['issuer'] = issuer

    try:
        headers = jwt.get_unverified_header(token)
        options['algorithms'] = [headers['alg']]
        signing_key = SIGNING_KEYS[headers['kid']]

        parsed = jwt.decode(token, signing_key, **options)
        rtn = {
            'user': parsed.get('sub')
        }

        claims_namespace = SETTINGS['AUTH0_CLAIMS_NAMESPACE']

        for claim in ['roles', 'acl', 'email', 'name', 'nickname']:
            value = parsed.get(f'{claims_namespace}/{claim}')
            if value:
                rtn[claim] = value
        if 'roles' in rtn:
            rtn['roles'] = [role.lower() for role in rtn['roles']]
            if 'admin' in rtn['roles']:
                rtn['role'] = 'admin'

    except (jwt.ExpiredSignatureError,
            jwt.InvalidSignatureError,
            jwt.InvalidAudienceError,
            jwt.InvalidAlgorithmError) as ex:  # TODO: other jwt ex's?
        rtn = {'_issues': {'token': f'{ex}'}}
    except ValueError as ex:
        rtn = {'_issues': {'config': f'{ex} - Please contact support for assistance, quoting this message'}}
    except jwt.DecodeError as ex:
        # NOTE: this should never occur - only did during development of this handler
        rtn = {'_issues': {'auth_handler': f'{ex} - Please contact support for assistance, quoting this message'}}
    except Exception as ex:
        rtn = {'_issues': {'unknown': f'{ex} - Please contact support for assistance, quoting this message'}}

    return rtn


def bearer_challenge(**kwargs):
    request = kwargs.get('request')
    rtn = {}
    if request and (
        'Bearer' in request.headers.get('Authorization', '')
        or request.args.get('access_token')
    ):
        rtn['error'] = "invalid_token"

    return rtn
